package jp.co.sds_corp.tool;

import java.io.File;
import java.net.JarURLConnection;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.regex.Pattern;

import jp.co.sds_corp.util.log.CmnLogFactory;
import jp.co.sds_corp.util.log.CmnLogger;

public class ClassFinder<X>
{
	static final CmnLogger logger = CmnLogFactory.getLogger();
	ClassLoader classLoader;

	public ClassFinder() {
		classLoader = Thread.currentThread().getContextClassLoader();
	}

	public ClassFinder(ClassLoader loader) {
		this.classLoader = loader;
	}

	@SuppressWarnings("unchecked")
	public List<Class<X>> findClass(String rootPackageName, X...x) throws Exception {
		return findClass(rootPackageName, new String[0], x);
	}

	@SuppressWarnings("unchecked")
	public List<Class<X>> findClass(String rootPackageName, String[] excludes, X...x) throws Exception {
		Class<X> clazz = (Class<X> )x.getClass().getComponentType();
		return findClass(rootPackageName, excludes, clazz);
	}

	@SuppressWarnings("unchecked")
	public List<Class<X>> findClass(String rootPackageName, List<Pattern> excludes, X...x) throws Exception {
		Class<X> clazz = (Class<X> )x.getClass().getComponentType();
		return findClass(rootPackageName, excludes, clazz);
	}

	public <T> List<Class<T>> findClass(String rootPackageName, Class<T> clazz) throws Exception {
		return findClass(rootPackageName, new String[0], clazz);
	}

	public <T> List<Class<T>> findClass(String rootPackageName, String[] excludes, Class<T> clazz) throws Exception {
		List<Pattern> list = new ArrayList<Pattern>();
		for (String exclude : excludes) {
			list.add(Pattern.compile(exclude));
		}
		return findClass(rootPackageName, list, clazz);
	}

	public <T> List<Class<T>> findClass(String rootPackageName, List<Pattern> excludes, Class<T> clazz) throws Exception {
		logger.trace(this, "targetClass=" + clazz);
		List<Class<T>> list = new ArrayList<Class<T>>();
		String resource = packageToResource(rootPackageName);
		Enumeration<URL> urls = classLoader.getResources(resource);

		while (urls.hasMoreElements()) {
			URL url = urls.nextElement();
			logger.trace(this, "url=" + url);

			String protocol = url.getProtocol();
			if ("file".equals(protocol)) {
				list.addAll(listClass(rootPackageName, new File(url.getFile()), excludes, clazz));
			} else if ("jar".equals(protocol)) {
				list.addAll(listClass(rootPackageName, url, excludes, clazz));
			} else {
				throw new IllegalArgumentException("Unsupported Protocol[" + protocol + "]");

			}
		}
		return list;
	}

	@SuppressWarnings("unchecked")
	public <T> List<Class<T>> listClass(String packageName, File dir, List<Pattern> excludes, Class<T> clazz) throws Exception {
		List<Class<T>> classes = new ArrayList<Class<T>>();

		for (String path : dir.list()) {
			File entry = new File(dir, path);
			if (entry.isFile() && isClassFile(entry.getName())) {
				String className  = packageName + "." + fileToClass(entry.getName());
				if (matches(excludes, className)) {
					continue;
				}
				Class<?> cls = classLoader.loadClass(className);
				if (clazz.isAssignableFrom(cls)) {
					classes.add((Class<T> )cls);
				}
			} else if (entry.isDirectory()) {
				classes.addAll(listClass(packageName + "." + entry.getName(), entry, excludes, clazz));
			}
		}

		return classes;
	}

	@SuppressWarnings("unchecked")
	public <T> List<Class<T>> listClass(String rootPackageName, URL jarFileUrl, List<Pattern> excludes, Class<T> clazz) throws Exception {
		List<Class<T>> classes = new ArrayList<Class<T>>();

		JarURLConnection jarUrlConnection = (JarURLConnection)jarFileUrl.openConnection();
		JarFile jarFile = null;

		try {
			jarFile = jarUrlConnection.getJarFile();
			Enumeration<JarEntry> en = jarFile.entries();

			String resource = packageToResource(rootPackageName);

			while (en.hasMoreElements()) {
				JarEntry entry = en.nextElement();
				if (entry.getName().startsWith(resource) && isClassFile(entry.getName())) {
					String className  = resourceToClass(entry.getName());
					if (matches(excludes, className)) {
						continue;
					}
					Class<?> cls = classLoader.loadClass(className);
					if (clazz.isAssignableFrom(cls)) {
						classes.add((Class<T> )cls);
					}
				}
			}
		} finally {
			if (jarFile != null) {
				jarFile.close();
			}
		}

		return classes;
	}

	boolean matches(List<Pattern> list, String name) {
		for (Pattern pattern : list) {
			if (pattern.matcher(name).matches()) {
				return true;
			}
		}
		return false;
	}

	String fileToClass(String name) {
		return name.substring(0, name.length() - ".class".length());
	}

	String resourceToClass(String resourceName) {
		return fileToClass(resourceName).replace('/', '.');
	}

	boolean isClassFile(String fileName) {
		return fileName.endsWith(".class");
	}

	String packageToResource(String packageName) {
		return packageName.replace('.', '/');
	}

}
